from numpy import *
from numba import cuda
import numba
import math, cmath

##################### not yet correct ###############################

@cuda.jit(numba.void(numba.complex128[:, :], numba.complex128[:, :], numba.complex128[:, :]),
          device=True)
def dot4(A, B, D):
    D[0, 0]=(A[0, 0]*B[0, 0]+A[0, 1]*B[1, 0]+A[0, 2]*B[2, 0]+
             A[0, 3]*B[3, 0])
    D[0, 1]=(A[0, 0]*B[0, 1]+A[0, 1]*B[1, 1]+A[0, 2]*B[2, 1]+
             A[0, 3]*B[3, 1])
    D[0, 2]=(A[0, 0]*B[0, 2]+A[0, 1]*B[1, 2]+A[0, 2]*B[2, 2]+
             A[0, 3]*B[3, 2])
    D[0, 3]=(A[0, 0]*B[0, 3]+A[0, 1]*B[1, 3]+A[0, 2]*B[2, 3]+
             A[0, 3]*B[3, 3])

    D[1, 0]=(A[1, 0]*B[0, 0]+A[1, 1]*B[1, 0]+A[1, 2]*B[2, 0]+
             A[1, 3]*B[3, 0])
    D[1, 1]=(A[1, 0]*B[0, 1]+A[1, 1]*B[1, 1]+A[1, 2]*B[2, 1]+
             A[1, 3]*B[3, 1])
    D[1, 2]=(A[1, 0]*B[0, 2]+A[1, 1]*B[1, 2]+A[1, 2]*B[2, 2]+
             A[1, 3]*B[3, 2])
    D[1, 3]=(A[1, 0]*B[0, 3]+A[1, 1]*B[1, 3]+A[1, 2]*B[2, 3]+
             A[1, 3]*B[3, 3])

    D[2, 0]=(A[2, 0]*B[0, 0]+A[2, 1]*B[1, 0]+A[2, 2]*B[2, 0]+
             A[2, 3]*B[3, 0])
    D[2, 1]=(A[2, 0]*B[0, 1]+A[2, 1]*B[1, 1]+A[2, 2]*B[2, 1]+
             A[2, 3]*B[3, 1])
    D[2, 2]=(A[2, 0]*B[0, 2]+A[2, 1]*B[1, 2]+A[2, 2]*B[2, 2]+
             A[2, 3]*B[3, 2])
    D[2, 3]=(A[2, 0]*B[0, 3]+A[2, 1]*B[1, 3]+A[2, 2]*B[2, 3]+
             A[2, 3]*B[3, 3])

    D[3, 0]=(A[3, 0]*B[0, 0]+A[3, 1]*B[1, 0]+A[3, 2]*B[2, 0]+
             A[3, 3]*B[3, 0])
    D[3, 1]=(A[3, 0]*B[0, 1]+A[3, 1]*B[1, 1]+A[3, 2]*B[2, 1]+
             A[3, 3]*B[3, 1])
    D[3, 2]=(A[3, 0]*B[0, 2]+A[3, 1]*B[1, 2]+A[3, 2]*B[2, 2]+
             A[3, 3]*B[3, 2])
    D[3, 3]=(A[3, 0]*B[0, 3]+A[3, 1]*B[1, 3]+A[3, 2]*B[2, 3]+
             A[3, 3]*B[3, 3])

@cuda.jit(numba.void(numba.float64[:], numba.complex128[:], numba.complex128[:],
                     numba.float64[:], numba.float64[:], numba.float64[:],
                     numba.float64[:, :], numba.complex128[:, :, :, :]))
def ReflNBSigma(Q, Vp, Vm, d, M_ang, sigma, Rout, PX):
    '''A quicker implementation than the ordinary slow implementaion in Refl
    Calculates spin-polarized reflectivity according to S.J. Blundell
        and J.A.C. Bland Phys rev. B. vol 46 3391 (1992)
        The algorithm assumes that the first element in the arrays represents
        the substrate and the last the ambient layer.
        Input parameters:   Q : Scattering vector in reciprocal
                                angstroms Q=4*pi/lambda *sin(theta)
                            Vp: Neutron potential for spin up
                            Vm: Neutron potential for spin down
                            d: layer thickness
                            M_ang: Angle of the magnetic
                                    moment(radians!) M_ang=0 =>M//neutron spin
                            sigma: The roughness of the upper interface.
                            return_int: Flag for returning the instensity, default=True. If False return the amplitudes.
        Returns:            (Ruu,Rdd,Rud,Rdu)
                            (up-up,down-down,up-down,down-up)
    '''
    layers=Vp.shape[0]

    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    qi=tx+ty*bw
    if qi>=Q.shape[0]:
        return

    X=cuda.local.array((4, 4), dtype=numba.complex128)
    P=cuda.local.array((4,), dtype=numba.complex128)
    M=cuda.local.array((4, 4), dtype=numba.complex128)
    Mtmp=cuda.local.array((4, 4), dtype=numba.complex128)

    # Assume first element=substrate and last=ambient!
    k_amb2=(Q[qi]/2.0)**2

    k_pi=cmath.sqrt(k_amb2-Vp[0])
    k_mi=cmath.sqrt(k_amb2-Vm[0])

    for lj in range(1, layers):
        # Wavevectors in the layers
        k_pj=cmath.sqrt(k_amb2-Vp[lj])
        k_mj=cmath.sqrt(k_amb2-Vm[lj])

        theta_diff=M_ang[lj]-M_ang[lj-1]

        ##### ass_X ####
        # Angular difference between the magnetization
        # Assemble the interface reflectivity matrix
        costd=math.cos(theta_diff/2.0)
        sintd=math.sin(theta_diff/2.0)
        X[0, 0]=costd*(k_pi+k_pj)/2./k_pj
        X[0, 1]=-costd*(k_pi-k_pj)/2./k_pj
        X[0, 2]=sintd*(k_pj+k_mi)/2./k_pj
        X[0, 3]=sintd*(k_pj-k_mi)/2./k_pj
        X[1, 0]=X[0, 1]  # -(costd*(k_pj1 - k_pj))/(2*k_pj)
        X[1, 1]=X[0, 0]  # (costd*(k_pj1 + k_pj))/(2*k_pj)
        X[1, 2]=X[0, 3]  # (sintd*(k_pj - k_mj1))/(2*k_pj)
        X[1, 3]=X[0, 2]  # (sintd*(k_pj + k_mj1))/(2*k_pj)
        X[2, 0]=-(sintd*(k_pi+k_mj))/(2.*k_mj)
        X[2, 1]=(sintd*(k_pi-k_mj))/(2.*k_mj)
        X[2, 2]=(costd*(k_mi+k_mj))/(2.*k_mj)
        X[2, 3]=-(costd*(k_mi-k_mj))/(2.*k_mj)
        X[3, 0]=X[2, 1]  # (sintd*(k_pj1 - k_mj))/(2*k_mj)
        X[3, 1]=X[2, 0]  # -(sintd*(k_pj1 + k_mj))/(2*k_mj)
        X[3, 2]=X[2, 3]  # -(costd*(k_mj1 - k_mj))/(2*k_mj)
        X[3, 3]=X[2, 2]  # (costd*(k_mj1 + k_mj))/(2*k_mj)

        ##### include_sigma #####
        sigma2=sigma[lj]**2/2.0
        X[0, 0]=X[0, 0]*cmath.exp(-(k_pj-k_pi)**2*sigma2)
        X[0, 1]=X[0, 1]*cmath.exp(-(k_pj+k_pi)**2*sigma2)
        X[0, 2]=X[0, 2]*cmath.exp(-(k_pj-k_mi)**2*sigma2)
        X[0, 3]=X[0, 3]*cmath.exp(-(k_pj+k_mi)**2*sigma2)
        X[1, 0]=X[0, 1]  # X[1,0]*w(k_pj + k_pj1, sigma2)
        X[1, 1]=X[0, 0]  # X[1,1]*w(k_pj - k_pj1, sigma2)
        X[1, 2]=X[0, 3]  # X[1,2]*w(k_pj + k_mj1, sigma2)
        X[1, 3]=X[0, 2]  # X[1,3]*w(k_pj - k_mj1, sigma2)
        X[2, 0]=X[2, 0]*cmath.exp(-(k_mj-k_pi)**2*sigma2)
        X[2, 1]=X[2, 1]*cmath.exp(-(k_mj+k_pi)**2*sigma2)
        X[2, 2]=X[2, 2]*cmath.exp(-(k_mj-k_mi)**2*sigma2)
        X[2, 3]=X[2, 3]*cmath.exp(-(k_mj+k_mi)**2*sigma2)
        X[3, 0]=X[2, 1]  # X[3,0]*w(k_mj + k_pj1, sigma)
        X[3, 1]=X[2, 0]  # X[3,1]*w(k_mj - k_pj1, sigma)
        X[3, 2]=X[2, 3]  # X[3,2]*w(k_mj + k_mj1, sigma)
        X[3, 3]=X[2, 2]  # X[3,3]*w(k_mj - k_mj1, sigma)

        ##### ass_P ####
        P[0]=cmath.exp(-1.0J*k_pj*d[lj])
        P[1]=1./P[0]  # exp(1.0J*k_p*d)
        P[2]=cmath.exp(-1.0J*k_mj*d[lj])
        P[3]=1./P[2]  # exp(1.0J*k_m*d)

        # Assemble the layer propagation matrices

        # Multiply the propagation matrices with the interface matrix
        for i in range(4):
            for j in range(4):
                PX[qi, lj, i, j]=P[i]*X[i, j]
        k_pi=k_pj
        k_mi=k_mj

    ##### ass_P ####
    for i in range(4):
        for j in range(4):
            M[i, j]=PX[qi, -2, i, j]
    for linv in range(layers-3):
        # Multiply up the sample matrix
        dot4(M, PX[qi, layers-3-linv], Mtmp)
        for i in range(4):
            for j in range(4):
                M[i, j]=Mtmp[i, j]
    dot4(X, M, Mtmp)
    for i in range(4):
        for j in range(4):
            M[i, j]=Mtmp[i, j]

    denom=M[0, 0]*M[2, 2]-M[0, 2]*M[2, 0]
    Ruu=(M[1, 0]*M[2, 2]-M[1, 2]*M[2, 0])/denom
    Rud=(M[3, 0]*M[2, 2]-M[3, 2]*M[2, 0])/denom
    Rdu=(M[1, 2]*M[0, 0]-M[1, 0]*M[0, 2])/denom
    Rdd=(M[3, 2]*M[0, 0]-M[3, 0]*M[0, 2])/denom
    Rout[0, qi]=abs(Ruu)**2
    Rout[1, qi]=abs(Rdd)**2
    Rout[2, qi]=abs(Rud)**2
    Rout[3, qi]=abs(Rdu)**2

@cuda.jit(numba.void(numba.float64[:], numba.complex128[:], numba.complex128[:],
                     numba.float64[:], numba.float64[:],
                     numba.float64[:, :], numba.complex128[:, :, :, :]))
def ReflNB(Q, Vp, Vm, d, M_ang, Rout, PX):
    '''A quicker implementation than the ordinary slow implementaion in Refl
    Calculates spin-polarized reflectivity according to S.J. Blundell
        and J.A.C. Bland Phys rev. B. vol 46 3391 (1992)
        The algorithm assumes that the first element in the arrays represents
        the substrate and the last the ambient layer.
        Input parameters:   Q : Scattering vector in reciprocal
                                angstroms Q=4*pi/lambda *sin(theta)
                            Vp: Neutron potential for spin up
                            Vm: Neutron potential for spin down
                            d: layer thickness
                            M_ang: Angle of the magnetic
                                    moment(radians!) M_ang=0 =>M//neutron spin
                            sigma: The roughness of the upper interface.
                            return_int: Flag for returning the instensity, default=True. If False return the amplitudes.
        Returns:            (Ruu,Rdd,Rud,Rdu)
                            (up-up,down-down,up-down,down-up)
    '''
    layers=Vp.shape[0]

    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    qi=tx+ty*bw
    if qi>=Q.shape[0]:
        return

    X=cuda.local.array((4, 4), dtype=numba.complex128)
    P=cuda.local.array((4,), dtype=numba.complex128)
    M=cuda.local.array((4, 4), dtype=numba.complex128)
    Mtmp=cuda.local.array((4, 4), dtype=numba.complex128)

    # Assume first element=substrate and last=ambient!
    k_amb2=(Q[qi]/2.0)**2

    k_pi=cmath.sqrt(k_amb2-Vp[0])
    k_mi=cmath.sqrt(k_amb2-Vm[0])

    for lj in range(1, layers):
        # Wavevectors in the layers
        k_pj=cmath.sqrt(k_amb2-Vp[lj])
        k_mj=cmath.sqrt(k_amb2-Vm[lj])

        theta_diff=M_ang[lj]-M_ang[lj-1]

        ##### ass_X ####
        # Angular difference between the magnetization
        # Assemble the interface reflectivity matrix
        costd=math.cos(theta_diff/2.0)
        sintd=math.sin(theta_diff/2.0)
        X[0, 0]=costd*(k_pi+k_pj)/2./k_pj
        X[0, 1]=-costd*(k_pi-k_pj)/2./k_pj
        X[0, 2]=sintd*(k_pj+k_mi)/2./k_pj
        X[0, 3]=sintd*(k_pj-k_mi)/2./k_pj
        X[1, 0]=X[0, 1]  # -(costd*(k_pj1 - k_pj))/(2*k_pj)
        X[1, 1]=X[0, 0]  # (costd*(k_pj1 + k_pj))/(2*k_pj)
        X[1, 2]=X[0, 3]  # (sintd*(k_pj - k_mj1))/(2*k_pj)
        X[1, 3]=X[0, 2]  # (sintd*(k_pj + k_mj1))/(2*k_pj)
        X[2, 0]=-(sintd*(k_pi+k_mj))/(2.*k_mj)
        X[2, 1]=(sintd*(k_pi-k_mj))/(2.*k_mj)
        X[2, 2]=(costd*(k_mi+k_mj))/(2.*k_mj)
        X[2, 3]=-(costd*(k_mi-k_mj))/(2.*k_mj)
        X[3, 0]=X[2, 1]  # (sintd*(k_pj1 - k_mj))/(2*k_mj)
        X[3, 1]=X[2, 0]  # -(sintd*(k_pj1 + k_mj))/(2*k_mj)
        X[3, 2]=X[2, 3]  # -(costd*(k_mj1 - k_mj))/(2*k_mj)
        X[3, 3]=X[2, 2]  # (costd*(k_mj1 + k_mj))/(2*k_mj)

        ##### ass_P ####
        P[0]=cmath.exp(-1.0J*k_pj*d[lj])
        P[1]=1./P[0]  # exp(1.0J*k_p*d)
        P[2]=cmath.exp(-1.0J*k_mj*d[lj])
        P[3]=1./P[2]  # exp(1.0J*k_m*d)

        # Assemble the layer propagation matrices

        # Multiply the propagation matrices with the interface matrix
        for i in range(4):
            for j in range(4):
                PX[qi, lj, i, j]=P[i]*X[i, j]
        k_pi=k_pj
        k_mi=k_mj

    ##### ass_P ####
    for i in range(4):
        for j in range(4):
            M[i, j]=PX[qi, -2, i, j]
    for linv in range(layers-3):
        # Multiply up the sample matrix
        dot4(M, PX[qi, layers-3-linv], Mtmp)
        for i in range(4):
            for j in range(4):
                M[i, j]=Mtmp[i, j]
    dot4(X, M, Mtmp)
    for i in range(4):
        for j in range(4):
            M[i, j]=Mtmp[i, j]

    denom=M[0, 0]*M[2, 2]-M[0, 2]*M[2, 0]
    Ruu=(M[1, 0]*M[2, 2]-M[1, 2]*M[2, 0])/denom
    Rud=(M[3, 0]*M[2, 2]-M[3, 2]*M[2, 0])/denom
    Rdu=(M[1, 2]*M[0, 0]-M[1, 0]*M[0, 2])/denom
    Rdd=(M[3, 2]*M[0, 0]-M[3, 0]*M[0, 2])/denom
    Rout[0, qi]=abs(Ruu)**2
    Rout[1, qi]=abs(Rdd)**2
    Rout[2, qi]=abs(Rud)**2
    Rout[3, qi]=abs(Rdu)**2

def Refl(Q, Vp, Vm, d, M_ang, sigma=None, return_int=True):
    threadsperblock=32
    blockspergrid=(Q.shape[0]+(threadsperblock-1))//threadsperblock

    CQ=cuda.to_device(Q)
    CVp=cuda.to_device(Vp.astype(complex128))
    CVm=cuda.to_device(Vm.astype(complex128))
    Cd=cuda.to_device(d)
    CM_ang=cuda.to_device(M_ang)
    CRout=cuda.device_array(shape=(4, Q.shape[0]), dtype=float64)
    CPX=cuda.device_array(shape=(Q.shape[0], Vp.shape[0], 4, 4), dtype=complex128)

    if sigma is not None:
        Csigma=cuda.to_device(sigma)
        ReflNBSigma[blockspergrid, threadsperblock](CQ, CVp, CVm, Cd, CM_ang, Csigma, CRout, CPX)
    else:
        ReflNB[blockspergrid, threadsperblock](CQ, CVp, CVm, Cd, CM_ang, CRout, CPX)

    return CRout.copy_to_host()



'''
Handling of a table of materials with associated density parameters.
Includes parsing of chemical formula string.
'''

import os
import re
import json

from genx.models.utils import UserVars, fw, fp, bc, __bc_dict__  # @UnusedImport
from genx.gui_logging import iprint

# configuration file to store the known materials
try:
    import appdirs
except ImportError:
    config_path=os.path.expanduser(os.path.join('~', '.genx'))
else:
    config_path=appdirs.user_data_dir('GenX3', 'ArturGlavic')
if not os.path.exists(config_path):
    os.makedirs(config_path)
config_file=os.path.join(config_path, 'materials.cfg')

default_materials=[
    [[["Cr", 1.0]], "7.19*0.602214/51.9961"],
    [[["D", 2.0], ["O", 1.0]], "1.107*0.602214/20.0276"],
    [[["Fe", 1.0]], "7.874*0.602214/55.845"],
    [[["Fe", 2.0], ["O", 3.0]], "2.0/100.713"],
    [[["H", 2.0], ["O", 1.0]], "1*0.602214/18.0152"],
    [[["La", 0.7], ["Sr", 0.3], ["Mn", 1.0], ["O", 3.0]], "6.0/349.916"],
    [[["Ni", 1.0]], "8.908*0.602214/58.6934"],
    [[["Si", 1.0]], "2.329*0.602214/28.0855"],
    [[["Si", 1.0], ["O", 2.0]], "3.0/113.005"],
    [[["Sr", 1.0], ["Ti", 1.0], ["O", 3.0]], "1.0/(3.905**3)"],
    [[["Ti", 1.0]], "4.506*0.602214/47.867"],
    [[["Ti", 1.0], ["O", 2.0]], "4.0/(4.5318*5.5019*4.9063)"],
    ]

MASS_DENSITY_CONVERSION=0.60221415  # g/cm³-> u/Å³ : 1e-24 (1/cm³->1/Å³) * 6.0221415e23 (Na)

class Formula(list):
    '''
    Holds the elements and fraction for a chemical formula.
    Includes the possibility to compare materials that
    have different element orders or base in their fraction.
    '''

    def __init__(self, data):
        # check that the data is correct form, list of [Element, fraction] items.
        for di in data:
            if len(di)!=2:
                raise ValueError('Formula has to consist of [Element, fraction] entries')
            if not self.check_atom(di[0]):
                raise ValueError('Element/Isotope %s no in database'%di[0])
            di[1]=float(di[1])
        list.__init__(self, data)

    @classmethod
    def from_str(cls, estr):
        '''
        Try to parse a formula string,
        has to be a simple formula with no brackets.
        
        Most read errors are ignored, so the result might not be
        the intended formula if the entry format is wrong.
        Format should be:
        {Element 1}{frac 1}{Element 2}{frac 2} like Fe2O3, SiO2 or H2O
        
        Fractions of 1 may be ommited and non-natural isotopes are
        entered with ^{N} before an element like ^{56}Fe2O3.
        '''
        if '(' in estr or ')' in estr:
            raise ValueError("Can't parse formula with brackets.")
        for ign_char in [" ", "\t", "_", "-"]:
            estr=estr.replace(ign_char, "")
        for i in range(10):
            estr=estr.replace(cls._get_subscript(i), '%i'%i)  # convert to normal str w/o subsccript
        if estr=="":
            return Formula([])
        extracted_elements=[]
        i=0
        mlen=len(estr)
        while i<mlen:
            # find next element
            iso_match=re.search('\^\{[0-9]{1,3}\}[A-Z][a-zA-Z]{0,1}', estr[i:])
            normal_match=re.search('[A-Z][a-zA-Z]{0,1}', estr[i:])
            if iso_match is None and normal_match is None:
                break
            elif iso_match is None or iso_match.start()>normal_match.start():
                element=estr[i+normal_match.start():i+normal_match.end()].capitalize()
            else:
                element=estr[i+iso_match.start():i+normal_match.start()]
                element+=estr[i+normal_match.start():i+normal_match.end()].capitalize()
            i+=normal_match.end()

            j=0
            while i+j<mlen and (estr[i+j]=='.' or estr[i+j].isdigit()):
                j+=1
            count_txt=estr[i:i+j]
            i+=j
            if count_txt=='':
                count=1.
            else:
                try:
                    count=float(count_txt)
                except ValueError:
                    continue
            extracted_elements.append([element, count])
        return Formula(extracted_elements)

    @classmethod
    def from_bstr(cls, bstr):
        '''
        Try to parse a string of scattering length elements,
        the same type that is returned by self.b().
        '''
        items=map(str.strip, bstr.split('+'))
        extracted_elements=[]
        for item in items:
            if not item.startswith('bc.'):
                continue
            else:
                element, count=item[3:].split('*', 1)
                count=float(count)
                extracted_elements.append([element, count])
        return Formula(extracted_elements)

    def __str__(self):
        '''Generate a string with sub- and superscript numbers for material.'''
        output=''
        for element, count in self:
            if element.startswith('^'):
                try:
                    isotope, element=element[2:].split('}')
                    output+=self._get_superscript(int(isotope))
                except (IndexError, ValueError):
                    pass
            if count==1:
                output+=element
            else:
                output+=element+self._get_subscript(count)
        return output

    def estr(self):
        '''Generates an editable (ascii) string that can be parsed back as Formul'''
        output=''
        for element, count in self:
            if count==1:
                output+=element
            else:
                output+=element+"%g"%count
        return output

    @classmethod
    def _get_superscript(cls, count):
        '''
          Return a subscript unicode string that equals the given number.
        '''
        scount='%g'%count
        result=''
        for char in scount:
            if char=='.':
                result+='﹒'
            else:
                # a superscript digit in unicode
                result+=(b'\\u207'+char.encode('utf-8')).decode('unicode-escape')
        return result

    @classmethod
    def _get_subscript(cls, count):
        '''
          Return a subscript unicode string that equals the given number.
        '''
        scount='%g'%count
        result=''
        for char in scount:
            if char=='.':
                result+='﹒'
            else:
                # a subscript digit in unicode
                result+=(b'\\u208'+char.encode('utf-8')).decode('unicode-escape')
        return result

    def elements(self):
        '''Returns alphabetically sorted list of elements in formula.'''
        return list(sorted([ei[0].split('}')[-1] for ei in self]))

    def isotopes(self):
        '''Returns same as elements but with isotopes w/ number at end of list.'''
        return list(sorted([ei[0] for ei in self]))

    def amounts(self):
        '''
        Amounts as stored by elements alphabetically.
        Can be used to compare two formulas for equality.
        '''
        items=[(ei[0].split('}')[-1], ei[1]) for ei in self]
        items.sort()
        return [i[1] for i in items]

    def fractions(self):
        '''
        Fractional numbers sorted by elements alphabetically.
        Can be used to compare two formulas for equivalency.
        '''
        fractions=self.amounts()
        total=sum(fractions)
        return [f/total for f in fractions]

    def __eq__(self, other):
        if type(other)!=Formula:
            try:
                cmpo=Formula(other)
            except:
                return False
        return self.elements()==other.elements() and self.amounts()==other.amounts()

    def equivalent(self, other):
        '''
        Returns if formula contains equivalent elemental composition,
        ignores the size of the fomula unit. For equality use ==.
        '''
        cmpo=Formula(other)
        return self.elements()==other.elements() and self.fractions()==other.fractions()

    def mFU(self):
        '''Calculate mass in u for formula unit (FU).'''
        mass=0.
        for ei, fi in self:
            try:
                mass+=fi*atomic_data[ei][2]
            except KeyError:
                if ei.startswith('^{'):
                    raise KeyError('Unknonw isotope: "%s"'%ei)
                else:
                    raise KeyError('Element %s does not exist'%ei)
        # return 1. if formulat is empty to avoid division by zero in density
        return mass or 1.0

    @staticmethod
    def check_atom(atom_string):
        return atom_string in atomic_data

    def describe(self):
        '''Return a multile string with written element content.'''
        output=''
        for ei, (element, number) in enumerate(self):
            if element.startswith('^'):
                iso='-%s'%(element[2:].split('}')[0])
                if not element in atomic_data:
                    iso='-unknown isotope'
                    element=element.split('}')[-1]
            else:
                iso=''
            output+="%g x %s\n"%(number, atomic_data[element][0]+iso)
        return output[:-1]

    def f(self):
        '''Return string to be used in models to calculate scattering power f'''
        if len(self)==0:
            return '0.j'
        fw.set_wavelength(1.54)
        elements=''
        for element, count in self:
            if element in isotopes:
                element=isotopes[element][1]
            elements+='+fp.%s*%g'%(element, count)
        return elements[1:]

    def b(self):
        '''Return string to be used in models to calculate scattering length b'''
        if len(self)==0:
            return '0.j'
        elements=''
        for element, count in self:
            if element in isotopes:
                element=isotopes[element][0]
            elements+='+bc.%s*%g'%(element, count)
        return elements[1:]

class MaterialsDatabase(list):
    '''
    Holds a list of materials and associated methods.
    '''

    def __init__(self):
        if os.path.exists(config_file):
            try:
                known_materials=json.loads(open(config_file, 'r').read())
            except json.JSONDecodeError:
                iprint("Can't reload material list, file corrupted.")
                known_materials=default_materials
        else:
            known_materials=default_materials
        data=[self.prepare(mi) for mi in known_materials]
        list.__init__(self, data)

    def save_data(self):
        open(config_file, 'w').write(json.dumps(self))

    def prepare(self, item):
        if len(item)!=2:
            raise ValueError("Requires [Formula, density_str] entry")
        return [Formula(item[0]), item[1]]

    def append(self, item):
        list.append(self, self.prepare(item))
        self.save_data()

    def __getitem__(self, item):
        if type(item) in [str, Formula]:
            # Try to match formula with database
            if type(item) is str:
                item=Formula.from_str(item)
            for material in self:
                if material[0]==item:
                    return material
                if material[0].equivalent(item):
                    material=[item, material[1]+'*%g'%(material[0].mFU()/item.mFU())]
                    return material
            raise IndexError("Material %s no in database"%item.estr())
        else:
            return list.__getitem__(self, item)

    def __contains__(self, item):
        if type(item) in [str, Formula]:
            # Try to match formula with database
            if type(item) is str:
                item=Formula.from_str(item)
            for material in self:
                if material[0]==item:
                    return True
                if material[0].equivalent(item):
                    return True
            return False
        else:
            return list.__contains__(self, item)

    def SLDx(self, item):
        density=self.dens_FU(item)
        fw.set_wavelength(1.54)
        return density*eval(self[item][0].f())  # *2.82

    def SLDn(self, item):
        density=self.dens_FU(item)
        return density*eval(self[item][0].b())*10.

    def dens_FU(self, item):
        '''Returns the formula unit (FU) density of the compound "item" in 1/Å³'''
        return eval(self[item][1])

    def dens_mass(self, item):
        '''Returns the mass density of the compound "item" in g/cm³'''
        return eval(self[item][1])*self[item][0].mFU()/MASS_DENSITY_CONVERSION

    def __delitem__(self, index):
        list.__delitem__(self, index)
        self.save_data()

    def __setitem__(self, index, item):
        list.__setitem__(index, self.prepare(item))
        self.save_data()

    def insert(self, index, item):
        list.insert(self, index, self.prepare(item))
        self.save_data()

    def pop(self, index):
        list.pop(self, index)
        self.save_data()

# list of elements with their name, atomic number and atomic mass values (+GenX name)
# mostly to calculate atomic density from mass density
atomic_data={
    "D": ("Deuterium", 1, 2.01410178),
    "H": ("Hydrogen", 1, 1.0079),
    "He": ("Helium", 2, 4.0026),
    "Li": ("Lithium", 3, 6.941),
    "Be": ("Beryllium", 4, 9.0122),
    "B": ("Boron", 5, 10.811),
    "C": ("Carbon", 6, 12.0107),
    "N": ("Nitrogen", 7, 14.0067),
    "O": ("Oxygen", 8, 15.9994),
    "F": ("Fluorine", 9, 18.9984),
    "Ne": ("Neon", 10, 20.1797),
    "Na": ("Sodium", 11, 22.9897),
    "Mg": ("Magnesium", 12, 24.305),
    "Al": ("Aluminum", 13, 26.9815),
    "Si": ("Silicon", 14, 28.0855),
    "P": ("Phosphorus", 15, 30.9738),
    "S": ("Sulfur", 16, 32.065),
    "Cl": ("Chlorine", 17, 35.453),
    "Ar": ("Argon", 18, 39.948),
    "K": ("Potassium", 19, 39.0983),
    "Ca": ("Calcium", 20, 40.078),
    "Sc": ("Scandium", 21, 44.9559),
    "Ti": ("Titanium", 22, 47.867),
    "V": ("Vanadium", 23, 50.9415),
    "Cr": ("Chromium", 24, 51.9961),
    "Mn": ("Manganese", 25, 54.938),
    "Fe": ("Iron", 26, 55.845),
    "Co": ("Cobalt", 27, 58.9332),
    "Ni": ("Nickel", 28, 58.6934),
    "Cu": ("Copper", 29, 63.546),
    "Zn": ("Zinc", 30, 65.39),
    "Ga": ("Gallium", 31, 69.723),
    "Ge": ("Germanium", 32, 72.64),
    "As": ("Arsenic", 33, 74.9216),
    "Se": ("Selenium", 34, 78.96),
    "Br": ("Bromine", 35, 79.904),
    "Kr": ("Krypton", 36, 83.8),
    "Rb": ("Rubidium", 37, 85.4678),
    "Sr": ("Strontium", 38, 87.62),
    "Y": ("Yttrium", 39, 88.9059),
    "Zr": ("Zirconium", 40, 91.224),
    "Nb": ("Niobium", 41, 92.9064),
    "Mo": ("Molybdenum", 42, 95.94),
    "Tc": ("Technetium", 43, 98),
    "Ru": ("Ruthenium", 44, 101.07),
    "Rh": ("Rhodium", 45, 102.906),
    "Pd": ("Palladium", 46, 106.42),
    "Ag": ("Silver", 47, 107.868),
    "Cd": ("Cadmium", 48, 112.411),
    "In": ("Indium", 49, 114.818),
    "Sn": ("Tin", 50, 118.71),
    "Sb": ("Antimony", 51, 121.76),
    "Te": ("Tellurium", 52, 127.6),
    "I": ("Iodine", 53, 126.904),
    "Xe": ("Xenon", 54, 131.293),
    "Cs": ("Cesium", 55, 132.905),
    "Ba": ("Barium", 56, 137.327),
    "La": ("Lanthanum", 57, 138.905),
    "Ce": ("Cerium", 58, 140.116),
    "Pr": ("Praseodymium", 59, 140.908),
    "Nd": ("Neodymium", 60, 144.24),
    "Pm": ("Promethium", 61, 145),
    "Sm": ("Samarium", 62, 150.36),
    "Eu": ("Europium", 63, 151.964),
    "Gd": ("Gadolinium", 64, 157.25),
    "Tb": ("Terbium", 65, 158.925),
    "Dy": ("Dysprosium", 66, 162.5),
    "Ho": ("Holmium", 67, 164.93),
    "Er": ("Erbium", 68, 167.259),
    "Tm": ("Thulium", 69, 168.934),
    "Yb": ("Ytterbium", 70, 173.04),
    "Lu": ("Lutetium", 71, 174.967),
    "Hf": ("Hafnium", 72, 178.49),
    "Ta": ("Tantalum", 73, 180.948),
    "W": ("Tungsten", 74, 183.84),
    "Re": ("Rhenium", 75, 186.207),
    "Os": ("Osmium", 76, 190.23),
    "Ir": ("Iridium", 77, 192.217),
    "Pt": ("Platinum", 78, 195.078),
    "Au": ("Gold", 79, 196.966),
    "Hg": ("Mercury", 80, 200.59),
    "Tl": ("Thallium", 81, 204.383),
    "Pb": ("Lead", 82, 207.2),
    "Bi": ("Bismuth", 83, 208.98),
    "Po": ("Polonium", 84, 209),
    "At": ("Astatine", 85, 210),
    "Rn": ("Radon", 86, 222),
    "Fr": ("Francium", 87, 223),
    "Ra": ("Radium", 88, 226),
    "Ac": ("Actinium", 89, 227),
    "Th": ("Thorium", 90, 232.038),
    "Pa": ("Protactinium", 91, 231.036),
    "U": ("Uranium", 92, 238.029),
    "Np": ("Neptunium", 93, 237),
    "Pu": ("Plutonium", 94, 244),
    "Am": ("Americium", 95, 243),
    "Cm": ("Curium", 96, 247),
    "Bk": ("Berkelium", 97, 247),
    "Cf": ("Californium", 98, 251),
    "Es": ("Einsteinium", 99, 252),
    "Fm": ("Fermium", 100, 257),
    "Md": ("Mendelevium", 101, 258),
    "No": ("Nobelium", 102, 259),
    "Lr": ("Lawrencium", 103, 262),
    "Rf": ("Rutherfordium", 104, 261),
    "Db": ("Dubnium", 105, 262),
    "Sg": ("Seaborgium", 106, 266),
    "Bh": ("Bohrium", 107, 264),
    "Hs": ("Hassium", 108, 277),
    "Mt": ("Meitnerium", 109, 268),
    }

isotopes={
    'D': ('i2H', 'H'),
    }

'''
    Go through the database of neutron scattering length and generate
    isotope list accordingly.
'''
for key in __bc_dict__.keys():
    if not key.startswith('i'):
        continue
    try:
        N=int(re.findall(r'\d+', key)[0])
    except IndexError:
        continue
    element=key[len('i%i'%N):].capitalize()
    isokey='^{%i}%s'%(N, element)
    n='i%i%s'%(N, element)
    isotopes[isokey]=(n, element)
    atomic_data[isokey]=(atomic_data[element][0],
                         atomic_data[element][1],
                         float(N))

mdb=MaterialsDatabase()



'''
Library that contains the the class Model. This
is a general class that store and binds togheter all the other
classes that is part of model parameters, such as data and parameters.
Programmer: Matts Bjorck
Last changed: 2016 09 20
'''

# Standard libraries
import shelve, os, types, zipfile
import pickle as pickle
import io, traceback
import inspect
import numpy as np
import h5py

# GenX libraries
from . import data
from . import parameters
from . import fom_funcs
from .gui_logging import iprint
from logging import debug

from .models.lib.parameters import NumericParameter, get_parameters

# assure compatibility with scripts that don't include genx in import
import sys

sys.modules['models']=sys.modules['genx.models']

# ==============================================================================
# BEGIN: Class Model

class Model:
    ''' A class that holds the model i.e. the script that defines
        the model and the data + various other attributes.
    '''

    def __init__(self, config=None):
        '''
        Create a instance and init all the varaibles.
        '''
        self.config=config
        self.data=data.DataList()
        self.script=''
        if config is not None:
            try:
                self.script="\n".join(eval(config.get('startup', 'script', fallback='[""]')))
            except:
                debug('Issue when loading script from config:', exc_info=True)
        self.parameters=parameters.Parameters(model=self)

        # self.fom_func = default_fom_func
        self.fom_func=fom_funcs.log  # The function that evaluates the fom
        self.fom=None  # The value of the fom function

        self.fom_mask_func=None
        self.fom_ignore_nan=False
        self.fom_ignore_inf=False
        self.create_fom_mask_func()

        # Registred classes that is looked for in the model
        self.registred_classes=[]
        # self.registred_classes = ['Layer','Stack','Sample','Instrument',\
        #                            'model.Layer', 'model.Stack',\
        #                             'model.Sample','model.Instrument',\
        #                             'UserVars','Surface','Bulk']
        self.set_func='set'  # 'set'
        self._reset_module()

        # Temporary stuff that needs to keep track on
        self.filename=''
        self.saved=True
        self.compiled=False

        self.limit_fit_range=False
        self.fit_xmin=0.01
        self.fit_xmax=0.1

        self.extra_analysis={}

    def read_config(self):
        '''Read in the config file
        '''
        # Ceck so that config is loaded
        if not self.config:
            return
        try:
            val=self.config.get('parameters', 'registred classes')
        except:
            iprint('Could not find config for parameters, registered classes')
        else:
            self.registred_classes=[s.strip() for s in val.split(';')]
        try:
            val=self.config.get('parameters', 'set func')
        except:
            iprint('Could not find config for parameters, set func')
        else:
            self.set_func=val

        try:
            val=self.config.get_boolean('solver', 'ignore fom nan')
        except:
            iprint('Could not find config for solver, ignore fom nan')
        else:
            self.fom_ignore_nan=val

        try:
            val=self.config.get_boolean('solver', 'ignore fom inf')
        except:
            iprint('Could not find config for solver, ignore fom inf')
        else:
            self.fom_ignore_inf=val

        self.create_fom_mask_func()

    def load(self, filename):
        ''' 
        Function to load the necessary parameters from a model file.
        '''
        try:
            loadfile=zipfile.ZipFile(filename, 'r')
        except Exception as e:
            raise IOError('Could not open file.', filename)
        try:
            new_data=pickle.loads(loadfile.read('data'), encoding='latin1', errors='ignore')
            self.data.safe_copy(new_data)
        except Exception as e:
            iprint('Data section loading (gx file) error:\n ', e, '\n')
            raise IOError('Could not locate the data section.', filename)
        try:
            self.script=pickle.loads(loadfile.read('script'), encoding='latin1', errors='ignore')
        except Exception as e:
            iprint('Script section loading (gx file) error:\n ', e, '\n')
            raise IOError('Could not locate the script.', filename)

        try:
            new_parameters=pickle.loads(loadfile.read('parameters'), encoding='latin1', errors='ignore')
            self.parameters.safe_copy(new_parameters)
        except Exception as e:
            iprint('Script section loading (gx file) error:\n ', e, '\n')
            raise IOError('Could not locate the parameters section.', filename)
        try:
            self.fom_func=pickle.loads(loadfile.read('fomfunction'), encoding='latin1', errors='ignore')
        except Exception:
            raise IOError('Could not locate the fomfunction section.', filename)

        loadfile.close()

        self.filename=os.path.abspath(filename)
        self.compiled=False
        self.saved=True
        self.script_module=types.ModuleType('genx_script_module')
        self.script_module.__dict__['data']=self.data
        self.script_module.__dict__['_sim']=False
        self.compiled=False

    def save(self, filename):
        '''
        Function to save the model to file filename
        '''
        try:
            savefile=zipfile.ZipFile(filename, 'w')
        except Exception as e:
            raise IOError(str(e), filename)

        # Save the data structures to file
        try:
            savefile.writestr('data', pickle.dumps(self.data))
        except Exception as e:
            raise IOError('Error writing data: '+str(e), filename)
        try:
            savefile.writestr('script', pickle.dumps(self.script))
        except Exception as e:
            raise IOError('Error writing script: '+str(e), filename)
        self.parameters.model=None
        try:
            savefile.writestr('parameters', pickle.dumps(self.parameters))
        except Exception as e:
            raise IOError('Error writing parameters: '+str(e), filename)
        try:
            savefile.writestr('fomfunction', pickle.dumps(self.fom_func))
        except Exception as e:
            raise IOError('Error writing fom_func:  '+str(e), filename)

        savefile.close()

        self.filename=os.path.abspath(filename)
        self.saved=True

    def write_h5group(self, group, **kwargs):
        """ Write the current parameters into a hdf5 group

        :param group: h5py Group to write into
        :param kwargs: Additional object parameters to be written (stored_name=object), these objects has to incorporate
            a write_h5group.
        :return:
        """
        self.data.write_h5group(group.create_group('data'))
        group['script']=self.script
        self.parameters.write_h5group(group.create_group('parameters'))
        group['fomfunction']=self.fom_func.__name__.encode('utf-8')
        sgrp=group.create_group('solver_pars')
        sgrp['fom_ignore_nan']=self.fom_ignore_nan
        sgrp['fom_ignore_inf']=self.fom_ignore_inf

        for kw in kwargs:
            kwargs[kw].write_h5group(group.create_group(kw))

    def read_h5group(self, group, **kwargs):
        """ Read the parameters from a hdf5 group

        :param group: hdf5 Group to read the parameters from
        :param kwargs: Additional object parameters to be read (stored_name=object), these objects has to incorporate
            a read_h5group.
        :return:
        """
        self.data.read_h5group(group['data'])
        if int(h5py.__version__[0])>2:
            self.script=group['script'][()].decode('utf-8')
        else:
            self.script=group['script'][()]
        self.parameters.read_h5group(group['parameters'])
        fom_func_name=group['fomfunction'][()].decode('utf-8')
        if fom_func_name in fom_funcs.func_names:
            self.set_fom_func(eval('fom_funcs.'+fom_func_name))
        else:
            iprint("Can not find fom function name %s"%fom_func_name[()])

        try:
            self.fom_ignore_nan=bool(group['solver_pars']['fom_ignore_nan'][()])
        except Exception as e:
            iprint("Could not load parameter fom_ignore_nan from file")
        try:
            self.fom_ignore_inf=bool(group['solver_pars']['fom_ignore_inf'][()])
        except Exception as e:
            iprint("Could not load parameter fom_ignore_inf from file")
        self.create_fom_mask_func()

        try:
            self.limit_fit_range, self.fit_xmin, self.fit_xmax=(
                bool(group['optimizer']['limit_fit_range'][()]),
                float(group['optimizer']['fit_xmin'][()]), float(group['optimizer']['fit_xmax'][()]))
        except Exception as e:
            iprint("Could not load limite_fit_range from file")

        for kw in kwargs:
            kwargs[kw].read_h5group(group[kw])

        self.compiled=False
        self.saved=True
        self.script_module=types.ModuleType('genx_script_module')
        self.script_module.__dict__['data']=self.data
        self.script_module.__dict__['_sim']=False
        self.compiled=False

    def save_addition(self, name, text):
        '''save_addition(self, name, text) --> None
        
        save additional text [string] subfile with name name [string]\
         to the current file.
        '''
        # Check so the filename is ok i.e. has been saved
        if self.filename=='':
            raise IOError('File must be saved before new information is added', '')
        try:
            savefile=zipfile.ZipFile(self.filename, 'a')
        except Exception as e:
            raise IOError(str(e), self.filename)

        # Check so the model data is not overwritten
        if name=='data' or name=='script' or name=='parameters':
            raise IOError('It not alllowed to save a subfile with name: %s'%name)

        try:
            savefile.writestr(name, text)
        except Exception as e:
            raise IOError(str(e), self.filename)
        savefile.close()

    def load_addition(self, name):
        '''load_addition(self, name) --> text
        
        load additional text [string] subfile with name name [string]\
         to the current model file.
        '''
        # Check so the filename is ok i.e. has been saved
        if self.filename=='':
            raise IOError('File must be loaded before additional information is read', '')
        try:
            loadfile=zipfile.ZipFile(self.filename, 'r')
        except Exception as e:
            raise IOError('Could not open the file', self.filename)

        try:
            text=loadfile.read(name)
        except Exception as e:
            raise IOError('Could not read the section named: %s'%name, self.filename)
        loadfile.close()
        return text

    def _reset_module(self):
        ''' 
        Internal method for resetting the module before compilation
        '''
        self.create_fom_mask_func()
        self.script_module=types.ModuleType('genx_script_module')
        # self.script_module = Temp()
        # self.script_module.__dict__ = {}
        # Bind data for preprocessing with the script
        self.script_module.__dict__['data']=self.data
        # Flag to indicate to the Sim funtion if a simulation is conducted (True)
        # or a fit is running (False). 
        self.script_module.__dict__['_sim']=False
        self.compiled=False

    def compile_script(self):
        ''' 
        compile the script in a seperate module.
        '''

        self._reset_module()
        # Testing to see if this works under windows
        self.script='\n'.join(self.script.splitlines())
        try:
            exec(self.script, self.script_module.__dict__)
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            raise ModelError(str(val), 0)
        else:
            self.compiled=True

    def eval_in_model(self, codestring):
        '''
        Excecute the code in codestring in the namespace of
        model module
        '''
        # exec codestring in self.script_module.__dict__
        result=eval(codestring, self.script_module.__dict__)
        # print 'Sucessfully evaluted: ', codestring
        return result

    def set_fom_ignore_inf(self, flag):
        """ Sets if the fom calculation should ignore infs

        :param flag: boolean
        :return:
        """
        self.fom_ignore_inf=bool(flag)
        self.create_fom_mask_func()

    def set_fom_ignore_nan(self, flag):
        """ Sets if fom calculations should ignore nan's

        :param flag: boolean flag
        :return:
        """
        self.fom_ignore_nan=bool(flag)
        self.create_fom_mask_func()

    def create_fom_mask_func(self):
        """ Create a mask func for fom to take care of unallowed values.

        :param fom: an array
        :return: an masked array
        """
        if self.fom_ignore_nan and self.fom_ignore_inf:
            fom_mask=lambda a: np.where(np.isfinite(a), a, np.zeros_like(a))
        elif self.fom_ignore_nan:
            fom_mask=lambda a: np.where(np.isnan(a), a, np.zeros_like(a))
        elif self.fom_ignore_inf:
            fom_mask=lambda a: np.where(np.isinf(a), a, np.zeros_like(a))
        else:
            fom_mask=lambda a: a*1.0

        self.fom_mask_func=fom_mask

    def calc_fom(self, simulated_data):
        '''calc_fom(self, fomlist) -> fom_raw (list of arrays), 
                                      fom_indiv(list of floats), 
                                      fom(float)
        
        Sums up the evaluation of the fom values calculated for each
         data point to form the overall fom function for all data sets.
        '''
        fom_raw=self.fom_func(simulated_data, self.data)
        # limit the x-range of fitting
        if self.limit_fit_range:
            for i, di in enumerate(self.data):
                fltr=(di.x<self.fit_xmin) | (di.x>self.fit_xmax)
                fom_raw[i][fltr]=0.
        # Sum up a unique fom for each data set in use
        fom_indiv=[np.sum(np.abs(self.fom_mask_func(fom_set))) for fom_set in fom_raw]
        fom=np.sum([f for f, d in zip(fom_indiv, self.data) if d.use])

        # Lets extract the number of datapoints as well:
        N=np.sum([len(self.fom_mask_func(fom_set)) for fom_set, d in zip(fom_raw, self.data) if d.use])
        # And the number of fit parameters
        p=self.parameters.get_len_fit_pars()
        # self.fom_dof = fom/((N-p)*1.0)
        try:
            use_dif=self.fom_func.__div_dof__
        except Exception:
            use_dif=False
        if use_dif:
            fom=fom/((N-p)*1.0)

        penalty_funcs=self.get_par_penalty()
        if len(penalty_funcs)>0 and fom is not np.NAN:
            fom+=sum([pf() for pf in penalty_funcs])
        return fom_raw, fom_indiv, fom

    def evaluate_fit_func(self):
        ''' evaluate_fit_func(self) --> fom (float)
        
        Evalute the Simulation fucntion and returns the fom. Use this one
        for fitting. Use evaluate_sim_func(self) for updating of plots
        and such.
        '''
        self.script_module._sim=False
        simulated_data=self.script_module.Sim(self.data)
        # fom = self.fom_func(simulated_data, self.data)
        fom_raw, fom_inidv, fom=self.calc_fom(simulated_data)
        return fom

    def evaluate_sim_func(self):
        '''evaluate_sim_func(self) --> None
        
        Evalute the Simulation function and updates the data simulated data
        as well as the fom of the model. Use this one for calculating data to
        update plots, simulations and such.
        '''
        self.script_module._sim=True
        try:
            simulated_data=self.script_module.Sim(self.data)
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            raise ModelError(str(val), 1)

        # check so that the Sim function returns anything
        if not simulated_data:
            text='The Sim function does not return anything, it should'+ \
                 ' return a list of the same length as the number of data sets.'
            raise ModelError(text, 1)
        # Check so the number of data sets is correct
        if len(simulated_data)!=len(self.data):
            text='The number of simulated data sets returned by the Sim function' \
                 +' has to be same as the number of loaded data sets.\n'+ \
                 'Number of loaded data sets: '+str(len(self.data))+ \
                 '\nNumber of simulated data sets: '+str(len(simulated_data))
            raise ModelError(text, 1)

        self.data.set_simulated_data(simulated_data)

        try:
            # self.fom = self.fom_func(simulated_data, self.data)
            fom_raw, fom_inidv, fom=self.calc_fom(simulated_data)
            self.fom=fom
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            raise FomError(str(val))
        # print len(fom_raw)
        self.data.set_fom_data(fom_raw)

    def create_fit_func(self, str):
        '''create_fit_func(self, str) --> function
        
        Creates a function from the string expression in string. 
        If the string is a function in the model this function will be
        returned if string represents anything else a function that sets that 
        object will be returned.
        '''
        object=self.eval_in_model(str)
        # Is it a function or a method!
        name=type(object).__name__
        if callable(object):
            return object
        # Make a function to set the object
        elif isinstance(object, NumericParameter):
            # We have a NumericParameter that should be set
            exec('def __tempfunc__(val):\n\t%s.value = val'%str, self.script_module.__dict__)
            return self.script_module.__tempfunc__
        else:
            # The function must be created in the module in order to access
            # the different variables
            exec('def __tempfunc__(val):\n\t%s = val'%str, self.script_module.__dict__)
            return self.script_module.__tempfunc__

    def get_fit_pars(self):
        ''' get_fit_pars(self) --> (funcs, values, min_values, max_values)
        
        Returns the parameters used with fitting. i.e. the function to 
        set the paraemters, the guess value (values), minimum allowed values
        and the maximum allowed values
        '''
        (row_numbers, sfuncs, vals, minvals, maxvals)=self.parameters.get_fit_pars()
        if len(sfuncs)==0:
            raise ParameterError(sfuncs, 0, 'None', 4)
        # Check for min and max on all the values
        for i in range(len(vals)):
            # parameter less than min
            if vals[i]<minvals[i]:
                raise ParameterError(sfuncs[i], row_numbers[i], 'None', 3)
            # parameter larger than max
            if vals[i]>maxvals[i]:
                raise ParameterError(sfuncs[i], row_numbers[i], 'None', 2)

        # Compile the strings to create the functions..
        funcs=[]
        for func in sfuncs:
            try:
                funcs.append(self.create_fit_func(func))
            except Exception as e:
                raise ParameterError(func, row_numbers[len(funcs)], str(e), 0)
        return funcs, vals, minvals, maxvals

    def get_par_penalty(self):
        for var in self.script_module.__dict__.values():
            if var.__class__.__name__=='UserVars':
                return var._penalty_funcs
        return []

    def get_fit_values(self):
        '''get_fit_values(self) --> values
        
        Returns the current parameters values that the user has ticked as
        fittable.
        '''
        (row_numbers, sfuncs, vals, minvals, maxvals)= \
            self.parameters.get_fit_pars()
        return vals

    def get_sim_pars(self):
        ''' get_sim_pars(self) --> (funcs, values)
        
        Returns the parameters used with simulations. i.e. the function to 
        set the parameters, the guess value (values). Used for simulation, 
        for fitting see get_fit_pars(self).s
        '''
        (sfuncs, vals)=self.parameters.get_sim_pars()
        # Compile the strings to create the functions..
        funcs=[]
        for func in sfuncs:
            try:
                funcs.append(self.create_fit_func(func))
            except Exception as e:
                raise ParameterError(func, len(funcs), str(e), 0)

        return funcs, vals

    def simulate(self, compile=True):
        '''simulate(self, compile = True) --> None
        
        Simulates the data sets using the values given in parameters...
        also compiles the script if asked for (default)
        '''
        if compile:
            self.compile_script()
        (funcs, vals)=self.get_sim_pars()
        # print 'Functions to evulate: ', funcs
        # Set the parameter values in the model
        # [func(val) for func,val in zip(funcs, vals)]
        i=0
        for func, val in zip(funcs, vals):
            try:
                func(val)
            except Exception as e:
                (sfuncs_tmp, vals_tmp)=self.parameters.get_sim_pars()
                raise ParameterError(sfuncs_tmp[i], i, str(e), 1)
            i+=1

        self.evaluate_sim_func()

    def new_model(self):
        '''
        new_model(self) --> None
        
        Reinitilizes the model. Thus, removes all the traces of the
        previous model. 
        '''
        iprint("class Model: new_model")
        self.data=data.DataList()
        self.script=''
        self.parameters=parameters.Parameters(self)

        # self.fom_func = default_fom_func
        self.fom_func=fom_funcs.log
        self._reset_module()

        # Temporary stuff that needs to keep track on
        self.filename=''
        self.saved=False

        self.extra_analysis={}

    def pickable_copy(self):
        '''pickable_copy(self) --> model
        
        Creates a pickable object of the model. Can be used for saving or
        sending to other processes, i.e., parallel processing.
        '''
        model_copy=Model(self.config)
        model_copy.data=self.data
        model_copy.script=self.script
        model_copy.parameters=self.parameters.copy()
        model_copy.fom_func=self.fom_func
        # The most important stuff - a module is not pickable
        model_copy.script_module=None
        model_copy.filename=self.filename
        model_copy.compiled=self.compiled
        model_copy.fom=self.fom
        model_copy.fom_ignore_nan=self.fom_ignore_nan
        model_copy.fom_ignore_inf=self.fom_ignore_inf
        model_copy.set_func=self.set_func
        model_copy.saved=self.saved
        # Needs to reset the fom_mask_func since this fails under windows.
        model_copy.fom_mask_func=None
        #
        model_copy.limit_fit_range=self.limit_fit_range
        model_copy.fit_xmin=self.fit_xmin
        model_copy.fit_xmax=self.fit_xmax

        return model_copy

    def get_table_as_ascii(self):
        '''get_table_as_ascii(self) --> None
        
        Just a copy of the parameters class method get_ascii_output()
        '''
        return self.parameters.get_ascii_output()

    def get_data_as_asciitable(self, indices=None):
        '''get_data_as_asciitable(self, indices None) --> string
        
        Just a copy of the method defined in data with the same name.
        '''
        return self.data.get_data_as_asciitable(indices)

    def export_table(self, filename):
        '''
        Export the table to filename. ASCII output.
        '''
        self._save_to_file(filename, self.parameters.get_ascii_output())

    def export_orso(self, basename):
        '''
        Export the data to files with basename filename. ORT output.
        The fileending will be .ort
        '''
        self.simulate(True)
        from genx.lib.orso_io import ort, data as odata
        from genx.version import __version__ as version
        para_list=[dict([(nj, tpj(pij)) for nj, pij, tpj in zip(self.parameters.data_labels, pi,
                                                           [str, float, bool, float, float, str])])
                   for pi in self.parameters.data if pi[0].strip()!='']
        add_header={'analysis': {
            'software': {'name': 'GenX', 'version': version},
            'script': self.script,
            'parameters': para_list,
            }}
        # add possible additional analysis data to the header
        add_header['analysis'].update(self.extra_analysis)
        ds=[]
        for di in self.data:
            header=dict(di.meta)
            try:
                import getpass
                header['creator']['name']=getpass.getuser()
            except Exception:
                pass
            import datetime
            header['creator']['time']=datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
            header.update(add_header)
            header['data_set']=di.name
            columns=[di.x, di.y, di.error]
            column_names=['Qz', 'R', 'sR']
            if 'res' in di.extra_data:
                columns.append(di.extra_data['res'])
                column_names.append('sQz')
            columns.append(di.y_sim)
            column_names.append('Rsim')
            columns.append(di.y_fom)
            column_names.append('FOM')
            for name, col in sorted(di.extra_data.items()):
                if name=='res':
                    continue
                columns.append(col)
                column_names.append(name)
            header['columns']=[{'name': cn} for cn in column_names]
            ds.append(odata.ORSOData(header, columns))
        try:
            ort.write_file(basename, ds)
        except data.IOError as e:
            raise IOError(e.error_message, e.file)

    def export_data(self, basename):
        '''
        Export the data to files with basename filename. ASCII output. 
        The fileending will be .dat
        First column is the x-values. 
        Second column is the data y-vales.
        Third column the error on the data y-values.
        Fourth column the calculated y-values.
        '''
        try:
            self.data.export_data_to_files(basename)
        except data.IOError as e:
            raise IOError(e.error_message, e.file)

    def export_script(self, filename):
        '''
        Export the script to filename. Will be a python script with ASCII 
        output (naturally).
        '''
        self._save_to_file(filename, self.script)

    def import_script(self, filename):
        '''import_script(self, filename) --> None
        
        Imports the script from file filename
        '''
        read_string=self._read_from_file(filename)
        self.set_script(read_string)
        self.compiled=False

    def import_table(self, filename):
        '''
        import the table from filename. ASCII input. tab delimited
        '''
        read_string=self._read_from_file(filename)
        self.parameters.set_ascii_input(read_string)

    def _save_to_file(self, filename, save_string):
        '''_save_to_file(self, filename, save_string) --> None
        
        Save the string to file with filename.
        '''
        try:
            savefile=open(filename, 'w')
        except Exception as e:
            raise IOError(e.__str__(), filename)

        # Save the string to file
        try:
            savefile.write(save_string)
        except Exception as e:
            raise IOError(e.__str__(), filename)

        savefile.close()

    def _read_from_file(self, filename):
        '''_read_from_file(self, filename) --> string
        
        Reads the entrie file into string and returns it.
        '''
        try:
            loadfile=open(filename, 'r')
        except Exception as e:
            raise IOError(e.__str__(), filename)

        # Read the text from file
        try:
            read_string=loadfile.read()
        except Exception as e:
            raise IOError(e.__str__(), filename)

        loadfile.close()

        return read_string

    # Get functions

    def get_parameters(self):
        '''
        get_parameters(self) --> parameters
        
        returns the parameters of the model. Instace of Parameters class
        '''
        return self.parameters

    def get_data(self):
        '''
        get_data(self) --> self.data
        
        Returns the DataList object.
        '''

        return self.data

    def get_script(self):
        '''
        get_script(self) --> self.script
        
        Returns the model script (string).
        '''
        return self.script

    def get_filename(self):
        '''
        get_filename(self) --> string
        
        returns the filename of the model id the model has not been saved
        it returns an empty string 
        '''
        return self.filename

    def get_possible_parameters(self):
        """ Returns all the parameters that can be fitted. Is used by the parameter grid.

        Returns:
             par_dict (list):  all the current objects that are of the classes defined by self.registred_classes.

        """
        # loop through all objects and locate NumericParameters or HasParameters
        par_dict=get_parameters(self.script_module, numeric_types_only=True)

        if len(par_dict)==0:
            par_dict=self.get_possible_set_functions()

        return par_dict

    def get_possible_set_functions(self):
        """Returns all the parameters that can be fitted given by the old style of defining parameters GenX2.4.X

        Returns:
             par_dict (list):  all the current objects that are of the classes defined by self.registred_classes.

        """
        # Start by updating the config file
        self.read_config()
        # First we should see if any of the 
        # classes is defnined in model.__pars__
        # or in __pars__
        pars=[]
        try:
            # Check if the have a pars in module named model
            pars_tmp=self.eval_in_model('model.__pars__')
            pars_tmp=['model.%s'%p for p in pars_tmp]
            pars+=pars_tmp
        except:
            pass

        # Check if we have a __pars__ in the main script
        try:
            pars_tmp=self.eval_in_model('__pars__')
            pars_tmp=['%s'%p for p in pars_tmp]
            pars+=pars_tmp
        except:
            pass

        isstrings=sum([type(p)==type('') for p in pars])==len(pars)

        if not isstrings:
            pars=[]

        # First find the classes that exists..
        # and defined in self.registred_classes
        classes=[]
        for c in self.registred_classes+pars:
            try:
                ctemp=self.eval_in_model(c)
            except:
                pass
            else:
                if inspect.isclass(ctemp):
                    classes.append(ctemp)
        # Check so there are any classes defined before we proceed
        if len(classes)>0:
            # Get all the objects in the compiled module
            names=list(self.script_module.__dict__.keys())
            # Create a tuple of the classes we defined above
            tuple_of_classes=tuple(classes)
            # Creating a dictionary that holds the name of the classes
            # eaxh item for a classes is a new dictonary that holds the
            # object name and then a list of the methods.
            par_dict={}
            [par_dict.__setitem__(clas.__name__, {}) for clas in classes]
            # find all the names of the objects that belongs to 
            # one of the classes
            objs=[(name, self.eval_in_model(name)) for name in names]
            valid_objs=[(name, obj) for name, obj in objs
                        if isinstance(obj, tuple_of_classes)]
            # nested for loop for finding for each valid object
            # the right name as given by self.set_func
            # Add this to the right item in par_dict given
            # its class and name.
            [par_dict[obj.__class__.__name__].__setitem__(name,
                                                          [member for member in dir(obj)
                                                           if member[:len(self.set_func)]==self.set_func])
             for name, obj in valid_objs]

            return par_dict

        return {}

    # Set functions
    def set_script(self, text):
        '''
        Set the text in the script use this to change the model script. 
        '''
        self.script=text
        self.compiled=False

    def set_fom_func(self, fom_func):
        '''
        Set the fucntion that calculates the figure of merit between the model
        and the data.
        '''
        self.fom_func=fom_func

    def is_compiled(self):
        '''is_compiled(self) --> compiled [boolean]
        
        Returns true if the model script has been sucessfully 
        compiled.
        '''
        return self.compiled

    def bumps_problem(self):
        """
        Return the fit function as a bumps problem to be used with the bumps.fitters.fit function.
        """
        if not self.compiled:
            self.compile_script()
        indep_module=self.script_module
        data=self.data
        # compile again to have a completely independent script module
        self.compile_script()

        from bumps.fitproblem import FitProblem
        from bumps.curve import Curve
        from numpy import hstack

        namespace={'script_module': indep_module, 'data': data,
                   'model_data': None, 'bmodel': None,
                   'Curve': Curve, 'hstack': hstack}

        # build function for full model parameters out of parameter grid
        mstring='def model_data(ignore'
        fstring=''
        pstart=[]
        for p in self.parameters:
            if p.name.strip()=='':
                continue
            name=p.name.replace('.set', '_')
            fstring+='script_module.%s(%s)\n'%(p.name, name)
            mstring+=', '+name
            pstart.append((name, p.value, p.min, p.max, p.fit))
        mstring+='):'
        fstring+='res=script_module.Sim(data)\nreturn hstack(res)'
        namespace['pstart']=pstart
        exec(mstring+'\n    '+fstring.replace('\n', '\n    '), namespace)
        exec('''bmodel=Curve(model_data, hstack([di.x for di in data]), 
                       hstack([di.y for di in data]), 
                       hstack([di.error for di in data]), 
                       '''+
             ', \n                   '.join(['%s=%s'%(pi[0], pi[1]) for pi in pstart])+')',
             namespace)
        for pname, pval, pmin, pmax, pfit in pstart:
            exec('bmodel.%s.range(%s, %s)'%(pname, pmin, pmax), namespace)
            if not pfit:
                exec('bmodel.%s.fixed=True'%pname, namespace)
        bproblem=FitProblem(namespace['bmodel'])
        return bproblem

    def bumps_fit(self, method='dream',
                  pop=15, samples=1e5, burn=100, steps=0,
                  thin=1, alpha=0, outliers='none', trim=False,
                  monitors=[], problem=None, **options):
        # create a fitter similar to the bumps.fitters.fit function but with option for GUI monitoring
        from scipy.optimize import OptimizeResult
        from bumps.fitters import FitDriver, FIT_AVAILABLE_IDS, FITTERS, FIT_ACTIVE_IDS
        options['pop']=pop
        options['samples']=samples
        options['burn']=burn
        options['steps']=steps
        options['thin']=thin
        options['alpha']=alpha
        options['outliers']=outliers
        options['trim']=trim

        if problem is None:
            problem=self.bumps_problem()

        # verbose = True
        if method not in FIT_AVAILABLE_IDS:
            raise ValueError("unknown method %r not one of %s"
                             %(method, ", ".join(sorted(FIT_ACTIVE_IDS))))
        for fitclass in FITTERS:
            if fitclass.id==method:
                break
        driver=FitDriver(fitclass=fitclass, problem=problem, monitors=monitors, **options)
        driver.clip()  # make sure fit starts within domain
        x0=problem.getp()
        x, fx=driver.fit()
        problem.setp(x)
        dx=driver.stderr()
        result=OptimizeResult(x=x, dx=driver.stderr(), fun=fx, cov=driver.cov(),
                              success=True, status=0, message="successful termination")
        if hasattr(driver.fitter, 'state'):
            result.state=driver.fitter.state
        return result

    def build_dream(self, x_n_gen=25, x_n_chain=5, sigmas=None):
        """
        Compile an independent script model and function to include in the
        Dream MCMC model and return that model.
        """
        if not self.compiled:
            self.compile_script()
        indep_module=self.script_module

        from bumps.dream import Dream
        from bumps.dream.model import Simulation
        from numpy import hstack, random

        data=self.data
        # compile again to have a completely independent script module
        self.compile_script()

        # generate a function that takes a single parameter list and returns the stacked data
        mstring='def model_data(p):\n    '
        fstring=''
        pstart=[]
        i=0
        for p in self.parameters:
            if not p.fit:
                continue
            name=p.name.replace('.set', '_')
            fstring+='script_module.%s(p[%i])\n'%(p.name, i)
            pstart.append((name, p.value, p.min, p.max))
            i+=1
        fstring+='res=script_module.Sim(data)\nreturn hstack(res)'
        namespace={'script_module': indep_module, 'data': data,
                   'model_data': None, 'hstack': hstack}
        exec(mstring+fstring.replace('\n', '\n    '), namespace)

        n=len(pstart)
        if sigmas is None:
            # if no ranges are supplied, use fit range from model
            bounds=(tuple(p[2] for p in pstart), tuple(p[3] for p in pstart))
        else:
            bounds=(tuple(p[1]+3*s[0] for p, s in zip(pstart, sigmas)),
                    tuple(p[1]+3*s[1] for p, s in zip(pstart, sigmas)))
        bsim=Simulation(f=namespace['model_data'],
                        data=hstack([di.y for di in data]),
                        sigma=hstack([di.error for di in data]),
                        bounds=bounds,
                        labels=[p[0] for p in pstart])
        bdream=Dream(model=bsim, draws=20000,
                     population=random.randn(x_n_gen*n, x_n_chain*n, n))  # n_gen, n_chain, n_var
        return bdream

    def plot(self, data_labels=None, sim_labels=None):
        """
        Plot all datasets in the model wtih matplotlib similar to the display in the GUI.
        data_labels and sim_labels are lists of strings for each curve,
        if none is supplied the curve label is created from the dataset names and index.
        """
        self.simulate()
        return self.data.plot(data_labels=data_labels, sim_labels=sim_labels)

    def __repr__(self):
        """
        Display information about the model.
        """
        output="Genx Model"
        if self.compiled:
            output+=' - compiled'
        else:
            output+=' - not compiled yet'
        output+="\n"
        output+="File: %s\n"%self.filename
        output+=self.parameters.__repr__()
        output+=self.data.__repr__()
        return output

    def _repr_html_(self):
        """
        Display information about the model.
        """
        try:
            self.simulate()
        except Exception as error:
            print(error)

        output="<h3>Genx Model"
        if self.compiled:
            output+=' - compiled'
        else:
            output+=' - not compiled yet'
        output+="</h3>\n"
        output+="<p>File: %s</p>\n"%self.filename

        output+='<div style="width: 100%;"><div style="width: 40%; float: left;">'
        output+=self.data._repr_html_()
        output+="</div>"

        # generate a plot of the model
        import binascii
        from io import BytesIO
        from matplotlib import pyplot as plt
        sio=BytesIO()
        fig=plt.figure(figsize=(10, 8))
        self.data.plot()
        plt.xlabel('q/tth')
        plt.ylabel('Intensity')
        fig.canvas.print_png(sio)
        plt.close()
        img_data=binascii.b2a_base64(sio.getvalue()).decode('ascii')
        output+='<div style="margin-left: 40%;"><img src="data:image/png;base64,{}&#10;"></div></div>'.format(img_data)

        output+=self.parameters._repr_html_()
        return output

    @property
    def widget(self):
        return self._repr_ipyw_()

    def _repr_ipyw_(self):
        import ipywidgets as ipw
        try:
            self.simulate()
        except Exception as error:
            print(error)

        graphw=ipw.Output()
        with graphw:
            from matplotlib import pyplot as plt
            fig=plt.figure(figsize=(10, 8))
            self.data.plot()
            plt.xlabel('q/tth')
            plt.ylabel('Intensity')
            from IPython.display import display
            display(fig)
            plt.close()

        dataw=self.data._repr_ipyw_()
        top=ipw.HBox([dataw, graphw])
        parameters=self.parameters._repr_ipyw_()

        replot=ipw.Button(description='simulate')
        replot.on_click(self._ipyw_replot)
        replot._plot_output=graphw

        script_area=ipw.Textarea(self.script, layout=ipw.Layout(width='100%'))
        script_area.observe(self._ipyw_script, names='value')
        tabs=ipw.Tab(children=[parameters, script_area])
        tabs.set_title(0, 'Parameters')
        tabs.set_title(1, 'Script')

        return ipw.VBox([replot, top, tabs])

    def _ipyw_replot(self, button):
        try:
            self.simulate()
        except Exception as error:
            print(error)
        with button._plot_output:
            from IPython.display import display, clear_output
            from matplotlib import pyplot as plt
            fig=plt.figure(figsize=(10, 8))
            self.data.plot()
            plt.xlabel('q/tth')
            plt.ylabel('Intensity')
            plt.draw()
            clear_output()
            display(fig)
            plt.close()

    def _ipyw_script(self, change):
        self.script=change.new

# END: Class Model
# ==============================================================================
# Some Exception definition for errorpassing
class GenericError(Exception):
    ''' Just a empty class used for inheritance. Only useful
    to check if the errors are originating from the model library.
    All these errors are controllable. If they not originate from
    this class something has passed trough and that should be impossible '''
    pass

class ParameterError(GenericError):
    ''' Class for yielding Parameter errors
    '''

    def __init__(self, parameter, parameter_number, error_message, what=-1):
        '''__init__(self, parameter, parameter_number, error_message) --> None
        
        parameter: the name of the parameter [string]
        parameter_number: the position of the parameter in the list [int]
        error_mesage: pythons error message from the original exception
        set: int to show where the error lies. 
            -1 : undefined
             0 : an not find the parameter
             1 : can not evaluate i.e. set the parameter
             2 : value are larger than max
             3 : value are smaller than min
             4 : No parameters to fit
        '''
        self.parameter=parameter
        self.parameter_number=parameter_number
        self.error_message=error_message
        self.what=what

    def __str__(self):
        ''' __str__(self) --> text [string]
        Yields a human readable description of the problem
        '''
        text=''
        text+='Parameter number %i, %s, '%(self.parameter_number,
                                           self.parameter)

        # Take care of the different cases
        if self.what==0:
            text+='could not be found. Check the spelling.\n'
        elif self.what==1:
            text+='could not be evaluated. Check the code of the function.\n'
        elif self.what==2:
            text+='is larger than the value in the max column.\n'
        elif self.what==3:
            text+='is smaller than the value in the min column\n'
        elif self.what==4:
            text='There are no parameter selcted to be fitted.\n'+ \
                 'Select the parameters you want to fit by checking the '+ \
                 'boxes in the fit column, folder grid'
        else:
            text+='yielded an undefined error. Check the Python output\n'

        if self.error_message!='None':
            text+='\nPython error output:\n'+self.error_message

        return text

class ModelError(GenericError):
    ''' Class for yielding compile or evaluation errors in the model text
    '''

    def __init__(self, error_message, where):
        '''__init__(self, error_message, where = -1) --> None
        
        error_mesage: pythons error message from the original exception
        where: integer describing where the error was raised.
                -1: undef
                 0: compile error
                 1: evaulation error
        '''
        self.error_message=error_message
        self.where=where

    def __str__(self):
        ''' __str__(self) --> text [string]
        Yields a human readable description of the problem
        '''
        text=''
        if self.where==0:
            text+='It was not possible to compile the model script.\n'
        elif self.where==1:
            text+='It was not possible to evaluate the model script.\n' \
                  +'Check the Sim function.\n'
        elif self.where==-1:
            text+='Undefined error from the Model. See below.\n'

        text+='\n'+self.error_message

        return text

class FomError(GenericError):
    '''Error class for the fom evaluation'''

    def __init__(self, error_message):
        ''' __init__(self, error_message) --> None'''
        self.error_message=error_message

    def __str__(self):
        text='Could not evaluate the FOM function. See python output.\n' \
             +'\n'+self.error_message
        return text

class IOError(GenericError):
    ''' Error class for input output, mostly concerning files'''

    def __init__(self, error_message, file=''):
        '''__init__(self, error_message)'''
        self.error_message=error_message
        self.file=file

    def __str__(self):
        text='Input/Output error for file:\n'+self.file+ \
             '\n\n Python error:\n '+self.error_message
        return text

# Some small default function that are needed for initilization

def default_fom_func(simulated_data, data):
    '''
    The default fom function. Its just a dummy so far dont use it!
    '''
    return sum([abs(d.y-sim_d).sum() for sim_d, d \
                in zip(simulated_data, data)])

class Temp:
    pass



#!/usr/bin/env python

import sys, os, appdirs, argparse
import os.path
from logging import debug

from . import version, model
from .gui_logging import setup_system, iprint, activate_logging, activate_excepthook

def start_interactive(args):
    ''' Start genx in interactive mode (with the gui)

    :param args:
    :return:
    '''
    debug('enter start_interactive')
    activate_excepthook()
    # Fix blurry text on Windows 10
    import ctypes
    try:
        ctypes.windll.shcore.SetProcessDpiAwareness(True)
    except:
        pass

    from genx import genx_gui
    if args.infile.endswith('.gx') or args.infile.endswith('.hgx'):
        debug('start GUI setup with file to load')
        app=genx_gui.MyApp(False, 0)
        # Create the window
        app.Yield()
        frame=app.TopWindow
        # load a model on start
        from genx import filehandling as io
        import traceback
        from io import StringIO
        from genx.event_handlers import ShowModelErrorDialog, ShowErrorDialog, get_pages, \
            _post_new_model_event, set_title
        from genx import model as modellib
        path=os.path.abspath(args.infile)
        try:
            io.load_file(path, frame.model, frame.solver_control.optimizer, frame.config)
        except modellib.IOError as e:
            ShowModelErrorDialog(frame, e.__str__())
        except Exception as e:
            outp=StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            ShowErrorDialog(frame, 'Could not open the file. Python Error:' \
                                   '\n%s'%(val,))
        else:
            app.Yield()
            try:
                [p.ReadConfig() for p in get_pages(frame)]
            except Exception as e:
                outp=StringIO()
                traceback.print_exc(200, outp)
                val=outp.getvalue()
                outp.close()
                ShowErrorDialog(frame, 'Could not read the config for the'
                                       ' plots. Python Error:\n%s'%(val,))
        # Letting the plugin do their stuff...
        try:
            frame.plugin_control.OnOpenModel(None)
        except Exception as e:
            outp=StringIO()
            traceback.print_exc(200, outp)
            val=outp.getvalue()
            outp.close()
            ShowErrorDialog(frame, 'Problems when plugins processed model.' \
                                   ' Python Error:\n%s'%(val,))
        frame.main_frame_statusbar.SetStatusText('Model loaded from file',
                                                 1)
        app.Yield()
        # Post an event to update everything else
        _post_new_model_event(frame, frame.model)
        # Needs to put it to saved since all the widgets will have
        # been updated
        frame.model.saved=True
        set_title(frame)
        app.Yield()
        # Just a force update of the data_list
        frame.data_list.list_ctrl.SetItemCount(frame.data_list.list_ctrl.data_cont.get_count())
        # Updating the imagelist as well
        frame.data_list.list_ctrl._UpdateImageList()
        frame.plot_data.plot_data(frame.model.data)
        frame.paramter_grid.SetParameters(frame.model.parameters)
        debug('setup complete, start WX MainLoop')
        app.MainLoop()
    elif args.infile=='':
        debug('start GUI setup')
        app=genx_gui.MyApp(True, 0)
        debug('setup complete, start WX MainLoop')
        app.MainLoop()
    else:
        iprint('Wrong file ending on infile, should be .gx or .hgx. Exiting.')
    debug('leave start_interactive')

def calc_errorbars(config, mod, opt):
    error_values=[]
    fom_error_bars_level=config.get_float('solver', 'errorbar level')
    n_elements=len(opt.start_guess)
    for index in range(n_elements):
        # calculate the error
        # TODO: Check the error bar buisness again and how to treat Chi2
        (error_low, error_high)=opt.calc_error_bar(index, fom_error_bars_level)
        error_values.append('(%.3e, %.3e,)'%(error_low, error_high))
    mod.parameters.set_error_pars(error_values)

def create_simulated_data(args):
    """Function to create simulated data from the model and add it to data.y"""

    from genx import model
    from genx import diffev
    from genx import filehandling as io

    from scipy.stats import poisson

    mod=model.Model()
    config=io.Config()
    config.load_default(os.path.split(os.path.abspath(__file__))[0]+'genx.conf')
    opt=diffev.DiffEv()

    iprint("Loading file: %s "%args.infile)
    io.load_file(args.infile, mod, opt, config)
    # io.load_opt_config(opt, config)
    iprint("File loaded")

    iprint("Simualting..")
    mod.simulate()
    iprint("Storing simualted data")
    for data_set in mod.data:
        data_set.y_raw=poisson.rvs(data_set.y_sim)
        data_set.y_command='y'
        data_set.run_y_command()
        data_set.error_command='sqrt(where(y > 0, y, 1.0))'
        data_set.run_error_command()

    iprint('Saving the model to %s'%args.outfile)
    io.save_file(args.outfile, mod, opt, config)

def extract_parameters(args):
    """Extracts the parameters to outfile"""

    from genx import model
    from genx import diffev
    from genx import filehandling as io

    # Open the genx file
    mod=model.Model()
    config=io.Config()
    config.load_default(os.path.split(os.path.abspath(__file__))[0]+'genx.conf')
    opt=diffev.DiffEv()

    iprint("Loading file: %s "%args.infile)
    io.load_file(args.infile, mod, opt, config)
    iprint("File loaded")

    names, values=mod.parameters.get_sim_pars()

    if args.outfile=='':
        outfile=sys.stdout
        fout=open(outfile, 'w')
    else:
        outfile=args.outfile
        if os.path.isfile(outfile):
            fout=open(outfile, 'a')
        else:
            fout=open(outfile, 'w')
            # Add header
            fout.write('\t'.join([name for name in names])+'\n')
    fout.write('\t'.join(['%f'%val for val in values])+'\n')
    fout.close()

def modify_file(args):
    """Modify a GenX file given command line arguments"""
    from genx import model
    from genx import diffev
    from genx import filehandling as io

    # Open the genx file
    mod=model.Model()
    config=io.Config()
    config.load_default(os.path.split(os.path.abspath(__file__))[0]+'genx.conf')
    opt=diffev.DiffEv()

    iprint("Loading file: %s "%args.infile)
    io.load_file(args.infile, mod, opt, config)
    iprint("File loaded")

    if args.datafile:
        datafile=os.path.abspath(args.datafile)
        if 0>args.data_set or args.data_set>=len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d"%(len(mod.data)-1))
            return
        iprint('Loading dataset %s into data set %d'%(datafile, args.data_set))
        mod.data[args.data_set].loadfile(datafile)

        if args.outfile:
            iprint('Saving the fit to %s'%args.outfile)
            io.save_file(args.outfile, mod, opt, config)

    elif args.save_datafile:
        save_datafile=os.path.abspath(args.save_datafile)
        iprint("Simualting..")
        mod.simulate()
        if 0>args.data_set or args.data_set>=len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d"%(len(mod.data)-1))
            return
        iprint('Exporting data set %d into ASCII file %s'%(args.data_set, save_datafile))
        mod.data[args.data_set].save_file(save_datafile)

def start_fitting(args, rank=0):
    """ Function to start fitting from the command line.

    :param args:
    :return:
    """
    import time
    from genx import model
    from genx import diffev
    from genx import filehandling as io

    mod=model.Model()
    config=io.Config()
    config.load_default(os.path.split(os.path.abspath(__file__))[0]+'genx.conf')
    opt=diffev.DiffEv()

    if rank==0:
        def autosave():
            # print 'Updating the parameters'
            mod.parameters.set_value_pars(opt.best_vec)
            if args.error:
                iprint("Calculating error bars")
                calc_errorbars(config, mod, opt)
            if args.outfile:
                iprint("Saving to %s"%args.outfile)
                io.save_file(args.outfile, mod, opt, config)

        opt.set_autosave_func(autosave)

    if rank==0:
        iprint('Loading model %s...'%args.infile)
    io.load_file(args.infile, mod, opt, config)
    io.load_opt_config(opt, config)
    # has to be used in order to save everything....
    if args.esave:
        config.set('solver', 'save all evals', True)
    # Simulate, this will also compile the model script
    if rank==0:
        iprint('Simulating model...')
    mod.simulate()

    # Sets up the fitting ...
    if rank==0:
        iprint('Setting up the optimizer...')
    set_optimiser_pars(opt, args)
    opt.reset()
    opt.init_fitting(mod)
    opt.init_fom_eval()
    opt.set_sleep_time(0.0)

    if args.outfile and rank==0:
        iprint('Saving the initial model to %s'%args.outfile)
        io.save_file(args.outfile, mod, opt, config)

    # To start the fitting
    if rank==0:
        iprint('Fitting starting...')
        t1=time.time()
    # print opt.use_mpi, opt.use_parallel_processing
    opt.optimize()
    if rank==0:
        t2=time.time()
        iprint('Fitting finished!')
        iprint('Time to fit: ', (t2-t1)/60., ' min')

    if rank==0:
        iprint('Updating the parameters')
        mod.parameters.set_value_pars(opt.best_vec)

    if args.outfile and rank==0:
        if args.error:
            iprint('Calculating errorbars')
            calc_errorbars(config, mod, opt)
        iprint('Saving the fit to %s'%args.outfile)
        opt.set_use_mpi(False)
        io.save_file(args.outfile, mod, opt, config)

    if rank==0:
        iprint('Fitting successfully completed')

def set_optimiser_pars(optimiser, args):
    """ Sets the optimiser parameters from args
    """
    if args.pr:
        optimiser.set_processes(args.pr)
        optimiser.set_use_parallel_processing(True)

    if args.mpi:
        optimiser.set_use_mpi(True)

    if args.cs:
        optimiser.set_chunksize(args.cs)

    if args.mgen:
        optimiser.set_max_generations(args.mgen)
        optimiser.set_use_max_generations(True)

    if args.pops:
        optimiser.set_pop_size(args.pops)
        optimiser.set_use_pop_mult(False)

    if args.asi:
        optimiser.set_autosave_interval(args.asi)
        optimiser.set_use_autosave(True)

    if args.km>=0:
        optimiser.set_km(args.km)
    # else:
    #    print "km not set has to be bigger than 0"

    if args.kr>=0:
        optimiser.set_kr(args.kr)
    # else:
    #    print "kr not set has to be bigger than 0"

def main():
    setup_system()
    import multiprocessing

    multiprocessing.freeze_support()
    # Attempt to load mpi:
    __mpi__=False
    try:
        from mpi4py import MPI
    except ImportError:
        pass
    else:
        __mpi__=True
        rank=MPI.COMM_WORLD.Get_rank()

    parser=argparse.ArgumentParser(description="GenX %s, fits data to a model."%version.__version__,
                                   epilog="For support, manuals and bug reporting see http://genx.sf.net"
                                   )
    run_group=parser.add_mutually_exclusive_group()
    run_group.add_argument('-r', '--run', action='store_true', help='run GenX fit (no gui)')
    if __mpi__:
        run_group.add_argument('--mpi', action='store_true', help='run GenX fit with mpi (no gui)')
    run_group.add_argument('-g', '--gen', action='store_true', help='generate data.y with poisson noise added')
    run_group.add_argument('--pars', action='store_true', help='extract the parameters from the infile')
    run_group.add_argument('--mod', action='store_true', help='modify the GenX file')
    opt_group=parser.add_argument_group('optimization arguments')
    opt_group.add_argument('--pr', type=int, default=0, help='Number of processes used in parallel fitting.')
    opt_group.add_argument('--cs', type=int, default=0, help='Chunk size used for parallel processing.')
    opt_group.add_argument('--mgen', type=int, default=0, help='Maximum number of generations that is used in a fit')
    opt_group.add_argument('--pops', type=int, default=0, help='Population size - number of individuals.')
    opt_group.add_argument('--asi', type=int, default=0, help='Auto save interval (generations).')
    opt_group.add_argument('--km', type=float, default=-1, help='Mutation constant (float 0 < km < 1)')
    opt_group.add_argument('--kr', type=float, default=-1, help='Cross over constant (float 0 < kr < 1)')
    opt_group.add_argument('-s', '--esave', action='store_true', help='Force save evals to gx file.')
    opt_group.add_argument('-e', '--error', action='store_true', help='Calculate error bars before saving to file.')
    data_group=parser.add_argument_group('data arguments')
    data_group.add_argument('-d', dest='data_set', type=int, default=0,
                            help='Active data set to act upon. Index starting at 0.')
    data_group.add_argument('--load', dest='datafile',
                            help='Load file into active data set. Index starting at 0.')
    data_group.add_argument('--export', dest='save_datafile',
                            help='Save active data set to file. Index starting at 0.')
    data_group.add_argument('-l', '--logfile', dest='logfile', default=None, type=str,
                            help='Output debug information to logfile.')
    data_group.add_argument('--debug', dest='debug', default=False, action="store_true",
                            help='Show additional debug information on console/logfile')

    parser.add_argument('infile', nargs='?', default='', help='The .gx or .hgx file to load')
    parser.add_argument('outfile', nargs='?', default='', help='The .gx  or hgx file to save into')

    args=parser.parse_args()
    if args.logfile:
        activate_logging(args.logfile)
    debug("Arguments from parser: %s"%args)
    if not args.outfile:
        args.outfile=args.infile
    args.outfile=os.path.abspath(args.outfile)
    if args.infile:
        args.infile=os.path.abspath(args.infile)
    path=os.path.split(model.__file__)[0]
    if os.path.abspath(path).endswith('.zip'):
        os.chdir(os.path.split(path)[0])
    else:
        os.chdir(path)

    if not __mpi__:
        args.mpi=False

    if args.run:
        start_fitting(args)
    elif args.mpi:
        start_fitting(args, rank)
    elif args.gen:
        create_simulated_data(args)
    elif args.pars:
        extract_parameters(args)
    elif args.mod:
        modify_file(args)
    elif not args.run and not args.mpi:
        # Check if the application has been frozen
        if hasattr(sys, "frozen") and True:
            # Redirect all the output to log files
            log_file_path=appdirs.user_log_dir('GenX3', 'ArturGlavic')
            # Create dir if not found
            if not os.path.exists(log_file_path):
                os.makedirs(log_file_path)
            # print log_file_path
            # log_file_path = genx_gui._path + 'app_data/'
            sys.stdout=open(log_file_path+'/genx.log', 'w')
            sys.stderr=open(log_file_path+'/genx.log', 'w')
        start_interactive(args)

if __name__=="__main__":
    main()


'''
========================================================
:mod:`auto` Data loader that selects plugin by extension
========================================================

Loads files using an other data loader that fits the file extension extracted from the file.
If none if found it falls back to the resolution loader if there res column is configured,
otherwise the default loader is used.

See the other data loaders for detailed explanation.
'''

from .resolution import Plugin as ResolutionPlugin
from .default import Plugin as DefaultPlugin
from .amor import Plugin as AmorPlugin
from .sns_mr import Plugin as SNSPlugin
from .d17_cosmos import Plugin as D17Plugin
from .orso import Plugin as ORSOPlugin

class Plugin(ResolutionPlugin, DefaultPlugin):
    """
    Automatic selected data loader.

    Plugins having fixed columns have to be put into the loaders list,
    if flexible columns are used they are part of the inheritance and can only use the
    columns defined in the ResolutionPlugin class.
    """

    def __init__(self, parent):
        ResolutionPlugin.__init__(self, parent)
        self.res_col=-1
        self.loaders=[AmorPlugin(None), SNSPlugin(None), D17Plugin(None), ORSOPlugin(None)]
        self.wildcard=";".join([li.wildcard for li in self.loaders])

    def LoadData(self, dataset, filename):
        for li in self.loaders:
            if li.CanOpen(filename):
                return li.LoadData(dataset, filename)
        if self.res_col<0:
            self.x_col=self.q_col
            self.y_col=self.I_col
            self.e_col=self.eI_col
            return DefaultPlugin.LoadData(self, dataset, filename)
        else:
            return ResolutionPlugin.LoadData(self, dataset, filename)
